import os
import time

start = time.time()

currentDirectory = os.getcwd()

print("working directory: " + currentDirectory)		
readpath = os.path.join(currentDirectory,'TransactionMaster','LIXI-Master-Schema.xsd')
print("read path: " + readpath)
with open(readpath, 'r') as file:
    data = file.read()
data = "var schemaString = `" + data + "`"
writepath = os.path.join(currentDirectory, 'js', 'LIXI-Master-Schema.js')
print("write path: " + writepath)
with open(writepath, 'w') as file:
    data = file.write(data)

    
readpath = os.path.join(currentDirectory,'TransactionMaster','LIXI_Glossary.xml')
print("read path: " + readpath)
with open(readpath, 'r') as file:
    data = file.read()
data = "var glossaryString = `" + data + "`"
writepath = os.path.join(currentDirectory, 'js', 'LIXI_Glossary.js')
print("write path: " + writepath)
with open(writepath, 'w') as file:
    data = file.write(data)


end = time.time()
print("\nComplete.")
print("   - Total Time:                      " + str(end - start))   



import json as _json
from lxml import etree as _etree
import re, io, os

ns = {
    "xs": "http://www.w3.org/2001/XMLSchema",
    "lx": "lixi.org.au/schema/appinfo_elements",
    "li": "lixi.org.au/schema/appinfo_instructions",
}

if __name__ == "__main__":
    import _xml_to_json, _path_functions, _jsonschema_functions, _xslt_transform, _schematron_functions
    from _LIXI import (
        LIXI as _LIXI,
        LIXIValidationError,
        LIXIInvalidSyntax,
        LIXIResouceNotFoundError,
    )
else:
    from lixi import (
        _xml_to_json,
        _path_functions,
        _jsonschema_functions,
        _xslt_transform,
        _schematron_functions,
    )
    from lixi._LIXI import (
        LIXI as _LIXI,
        LIXIValidationError,
        LIXIInvalidSyntax,
        LIXIResouceNotFoundError,
    )


class Message:
    """A LIXI message object (either XML or JSON) that conforms to a LIXI2 standard.
        
    The class exists as a wrapper for all internal message functions the library is to provide.
    """

    def __init__(self, data, message_path, file_type):
        """Creates a LIXI message XML.
    
        Args:
            data (:obj:`str`, optional): A LIXI Message provided as a string. Defaults to None.
            message_path (:obj:`str`, optional) : A LIXI Message provided as a path. Defaults to None.
            file_type (:obj:`str`, optional): The type of the input LIXI Message given. Defaults to 'xml'. DEPRECATED. But have kept it for later use if needed.
    
        Returns:
            A LIXI message instance.
    
        Raises:
            LIXIResouceNotFoundError: If the schema is not found at the schema path.
            LIXIInvalidSyntax: If the schema file is not well formed.
        """

        self.xml_package = None
        self.json_package = None

        self.json_string = None
        self.file_type = None

        self.schema_name = None
        self.is_valid = None
        self.validation_message = None

        # If data specified as source
        if data != None and message_path == None:
            # just a check to ensure 
            data = re.sub(r'\<\?.*?\?\>', '', data)
            
            passed = False
            if passed == False:
                try:
                    parser = _etree.XMLParser(remove_blank_text=True)
                    self.xml_package = _etree.fromstring(data, parser)
                    passed = True
                    self.file_type = "xml"
                except Exception as e:
                    passed = False

            if passed == False:
                try:
                    self.json_string = str(data)
                    self.json_package = _json.loads(data)

                    passed = True
                    self.file_type = "json"
                except Exception as e:
                    passed = False

        # If path specified as source
        elif data == None and message_path != None:
            passed = False
            
            if os.path.exists(message_path) == False:
                raise LIXIResouceNotFoundError(message_path)
            
            if passed == False:
                try:
                    self.xml_package = _etree.parse(message_path)
                    self.xml_package = self.xml_package.getroot()
                    passed = True
                    self.file_type = "xml"
                except Exception as e:
                    passed = False

            if passed == False:
                try:

                    with io.open(message_path, "r", encoding="utf-8") as json_file:
                        self.json_string = str(json_file.read())
                        self.json_package = _json.loads(str(self.json_string))

                    passed = True
                    self.file_type = "json"
                except Exception as e:
                    passed = False

        # Checking for some required fields
        if self.file_type == "xml":

            self.is_json = False

            try:
                self.lixi_transaction_type = self.xml_package.xpath(
                    "/Package/SchemaVersion/@LIXITransactionType"
                )[0]
            except IndexError:
                raise LIXIInvalidSyntax(
                    "Message read failed. XML Message does not have the attribute 'Package.SchemaVersion.LIXITransactionType' populated."
                )

            try:
                self.lixi_version = self.xml_package.xpath(
                    "/Package/SchemaVersion/@LIXIVersion"
                )[0]
            except IndexError:
                raise LIXIInvalidSyntax(
                    "Message read failed. XML Message does not have the attribute 'Package.SchemaVersion.LIXIVersion' populated."
                )

            try:
                self.lixi_custom_version = self.xml_package.xpath(
                    "/Package/SchemaVersion/@LIXICustomVersion"
                )[0]
            except IndexError:
                self.lixi_custom_version = None
            
            try:
                self.publisher_lixicode = self.xml_package.xpath(
                    "/Package/Publisher/@LIXICode"
                )[0]
            except IndexError:
                self.publisher_lixicode = None
                
            try:
                self.broker_reference = self.xml_package.xpath(
                    "/Package/Content/Application/Overview/@BrokerApplicationReferenceNumber"
                )[0]
            except IndexError:
                self.broker_reference = None
                
            try:
                self.lender_reference = self.xml_package.xpath(
                    "/Package/Content/Application/Overview/@LenderApplicationReferenceNumber"
                )[0]
            except IndexError:
                self.lender_reference = None
                    
            try:
                self.lodgement_reference = self.xml_package.xpath(
                    "/Package/Content/Application/Overview/@LodgementReferenceNumber"
                )[0]
            except IndexError:
                self.lodgement_reference = None
                
            try:
                self.valuation_requestor_assigned_id = self.xml_package.xpath(
                    "/Package/Content/RealEstateValuation/ValuationOrder/@RequestorAssignedID"
                )[0]
            except IndexError:
                self.valuation_requestor_assigned_id = None
                
            try:
                self.valuation_valuer_assigned_id = self.xml_package.xpath(
                    "/Package/Content/RealEstateValuation/@ValuerAssignedID"
                )[0]
            except IndexError:
                self.valuation_valuer_assigned_id = None
                
            try:
                self.valuation_avm_assigned_id = self.xml_package.xpath(
                    "/Package/Content/RealEstateValuation/@AVMReferenceID"
                )[0]
            except IndexError:
                self.valuation_avm_assigned_id = None                        

        elif self.file_type == "json":

            self.is_json = True

            try:
                self.lixi_transaction_type = str(
                    self.json_package["Package"]["SchemaVersion"][
                        "@LIXITransactionType"
                    ]
                )
            except Exception:
                raise LIXIInvalidSyntax(
                    "Message read failed. JSON Message does not have the attribute 'Package.SchemaVersion.LIXITransactionType' populated."
                )

            try:
                self.lixi_version = str(
                    self.json_package["Package"]["SchemaVersion"]["@LIXIVersion"]
                )
            except Exception:
                raise LIXIInvalidSyntax(
                    "Message read failed. JSON Message does not have the attribute 'Package.SchemaVersion.LIXIVersion' populated."
                )

            try:
                self.lixi_custom_version = str(
                    self.json_package["Package"]["SchemaVersion"]["@LIXICustomVersion"]
                )
            except Exception:
                self.lixi_custom_version = None
                
            try:
                self.publisher_lixicode = str(
                    self.json_package["Package"]["Publisher"]["@LIXICode"]
                )
            except Exception:
                self.publisher_lixicode = None
                
            try:
                self.broker_reference = str(
                    self.json_package["Package"]["Content"]["Application"]["Overview"]["@BrokerApplicationReferenceNumber"]
                )
            except Exception:
                self.broker_reference = None 
                
            try:
                self.lender_reference = str(
                    self.json_package["Package"]["Content"]["Application"]["Overview"]["@LenderApplicationReferenceNumbe"]
                )
            except Exception:
                self.lender_reference = None 
                
            try:
                self.lodgement_reference = str(
                    self.json_package["Package"]["Content"]["Application"]["Overview"]["@LodgementReferenceNumber"]
                )
            except Exception:
                self.lodgement_reference = None                            
            
            try:
                self.valuation_requestor_assigned_id = str(
                    self.json_package["Package"]["Content"]["RealEstateValuation"]["ValuationOrder"]["@RequestorAssignedID"]
                )
            except Exception:
                self.valuation_requestor_assigned_id = None 
                
            try:
                self.valuation_valuer_assigned_id = str(
                    self.json_package["Package"]["Content"]["RealEstateValuation"]["@ValuerAssignedID"]
                )
            except Exception:
                self.valuation_valuer_assigned_id = None 
                
            try:
                self.valuation_avm_assigned_id = str(
                    self.json_package["Package"]["Content"]["RealEstateValuation"]["@AVMReferenceID"]
                )
            except Exception:
                self.valuation_avm_assigned_id = None              
            
        else:  # everything failed now we need better error messages
            ##Better error messages
            error_message = ""

            if data == None and message_path != None:
                with io.open(message_path, "r", encoding="utf-8") as file:
                    data = str(file.read())

            if '"Package"' in data:  ## is json
                try:
                    _json.loads(str(data))
                except Exception as e:
                    error_message = e
            elif "<Package>" in data:
                try:
                    parser = _etree.XMLParser(remove_blank_text=True)
                    doc = _etree.fromstring(data, parser)
                except Exception as e:
                    error_message = e

            raise LIXIInvalidSyntax(
                "Message read failed. Message can not be read or type is not supported.\n"
                + str(error_message)
            )

    def __write__(self, default_name, output_path, data, error_string):
        """Helper function to write to a file.

        Args:
            default_name (:obj:`str`, required): A default name in case the provided out path is a folder path and not a complete file path.
            output_path (:obj:`str`, required): The output path of the file to be written.
            data (:obj:`str`, required): Data to be written to the  file.
            error_string (:obj:`str`, required): A context specific error string in case write fails.

        Raises:
            LIXIResouceNotFoundError: If output path provided does not exist.
        """

        if os.path.isdir(output_path):
            output_path = os.path.join(output_path, default_name)

        try:
            with io.open(output_path, "w+", encoding="utf-8") as out_file:
                out_file.write(data)
        except Exception as e:
            raise LIXIResouceNotFoundError(error_string)

    def to_json(self, return_instance=False):
        """Converts the LIXI XML Message object to a LIXI JSON Message object. Alternatively, returns a LIXI JSON Message object if return_instance parameter is set to True. 

        Args:
            return_instance (:obj:`bool`, required): Indicates if a LIXI message instance needs to be created from the result and returned. 

        Returns:
            Equivalent LIXI JSON message instance ONLY if return_instance is set to True. 

        Raises:
            LIXIInvalidSyntax: Message file is corrupted.
        """

        # for converstion an xsd is required to properly arrange elements.
        schema_etree = _LIXI.getInstance().fetch_xml_schema(
            lixi_transaction_type=self.lixi_transaction_type,
            lixi_version=self.lixi_version,
            custom_version=self.lixi_custom_version,
            return_annotated = True
        )
        schema_etree = _etree.fromstring(schema_etree)        

        if self.json_package == None:
            self.json_package = _xml_to_json.to_json(self.xml_package, schema_etree)

        if return_instance == True:
            return Message(
                _json.dumps(
                    self.json_package, indent=4, ensure_ascii=False
                ),
                None,
                "json",
            )
        else:
            self.file_type = "json"
            self.is_json = True

    def to_xml(self, return_instance=False):
        """Converts a LIXI JSON Message object to a LIXI XML Message object. Alternatively, returns a LIXI XML Message object if return_instance parameter is set to True. 


        Args:
            return_instance (:obj:`bool`, required): Indicates if a LIXI message instance needs to be created from the result and returned. 

        Returns:
            Equivalent LIXI XML message instance ONLY if return_instance is set to True. 

        Raises:
            LIXIInvalidSyntax: Message file is corrupted.
        """

        # for converstion an xsd is required hence setting it to be false
        schema_etree = _LIXI.getInstance().fetch_xml_schema(
            lixi_transaction_type=self.lixi_transaction_type,
            lixi_version=self.lixi_version,
            custom_version=self.lixi_custom_version,
            return_annotated = True
        )
        schema_etree = _etree.fromstring(schema_etree)

        if self.xml_package == None:
            self.xml_package = _xml_to_json.to_xml(self.json_package, schema_etree)

        if return_instance == True:
            return Message(
                _etree.tostring(self.xml_package, pretty_print=True).decode("utf-8"),
                None,
                "xml",
            )
        else:
            self.file_type = "xml"
            self.is_json = False

    def get_message_paths(self, output_path=None):
        """Derives all LIXI Paths (the full path of elements and attributes separated by dots, eg Package.Content.Application) within this LIXI Message object.

        Args:
            output_path (:obj:`str`, optional): Path to write the message paths list to. Defaults to None.

        Returns:
            A list of element paths in the Message. 

        Raises:
            LIXIResouceNotFoundError: If the output path provided is not correct.
        """

        # Schema to be used as per message to derive paths from
        schema = _LIXI.getInstance().fetch_xml_schema(
            lixi_transaction_type=self.lixi_transaction_type,
            lixi_version=self.lixi_version,
            custom_version=self.lixi_custom_version,
            return_annotated = True            
        )
        schema = _etree.fromstring(schema)

        # Incase this is a json message, functionality works with xml only
        if self.xml_package == None:
            self.xml_package = _xml_to_json.to_xml(self.json_package, schema)

        message_paths_list = _path_functions.get_paths_for_elements(
            message=self, schema=schema
        )

        # output to a path
        if output_path == None:
            return message_paths_list
        else:
            self.__write__(
                "message_paths_output.txt",
                output_path,
                ",\n".join(message_paths_list),
                "Message get paths failed. Can not store the schema at the specified folder.",
            )
            return message_paths_list

    def get_schema_paths(self, exploded=False, output_path=None):
        """Gets all the elements paths of the LIXI schema associated with this LIXI Message instance.
        Derives all LIXI Paths (the full path of elements and attributes separated by dots, eg Package.Content.Application) available in the schema corresponding to the this LIXI Message object.

        Args:
            exploded (:obj:`bool`, optional): Variable to indicate if the returned schema should be exploded or not. 
            output_path (:obj:`str`, optional): Path to write the schema paths list to. Defaults to None.

        Returns:
            A list of element paths in the LIXI schema. 

        Raises:
            LIXIResouceNotFoundError: If the output path provided is not correct.
        """

        # Gets the schema paths from LIXI function
        schema_paths_list = _LIXI.getInstance().get_schema_paths(
            self.lixi_transaction_type,
            self.lixi_version,
            "xml",
            self.lixi_custom_version,
            None,
            None,
            exploded
        )

        # Output paths
        if output_path == None:
            return schema_paths_list
        else:
            self.__write__(
                "schema_paths_output.txt",
                output_path,
                ",".join(schema_paths_list),
                "Message get schema paths failed. Can not store the schema at the specified folder.",
            )
            return schema_paths_list

    def get_restriction_paths_for_schema(self, output_path=None):
        """Gets the customization instructions file which can be used for a generating a customised schema based on the subset of elements and attributes that are actually included in this LIXI Message object.

        Args:
            output_path (:obj:`str`, optional): Path to write the schema paths list to. Defaults to None.

        Returns:
            A string of blacklist customization restrictions xml that can be utilized by this library to generate a custom schema. 

        Raises:
            LIXIResouceNotFoundError: If the output path provided is not correct.
        """

        message_paths = self.get_message_paths()
        schema_paths = self.get_schema_paths()

        customization_instructions = _path_functions.get_blacklist_paths_for_customization(
            self.lixi_transaction_type, message_paths, schema_paths
        )

        if output_path == None:
            return customization_instructions
        else:
            self.__write__(
                "customization_instructions_output.xml",
                output_path,
                customization_instructions,
                "Message get restriction paths failed. Can not store the schema at the specified folder.",
            )
            return customization_instructions

    def get_custom_schema(self, output_path=None):
        """Generates a schema using Customisation by Restriction based on the subset of elements and attributes that are actually included in this LIXI Message object.
    
        Args:
            output_path (:obj:`str`, optional): Path to write the schema paths list to. Defaults to None.
    
        Returns:
            A customised LIXI schema as a string which can also be to the output folder specified. 
    
        Raises:
            LIXIInvalidSyntax: Validation errors for the lixi version or transaction type.
            LIXIResouceNotFoundError: If path provided does not exist.
            LIXIValidationError: validation error in the customization instructions.
        """

        # Read the instructions file
        restriction_paths = self.get_restriction_paths_for_schema(output_path=None)

        # Generate the customised schema as string
        if output_path == None:
            return _LIXI.getInstance().generate_custom_schema(
                instructions=restriction_paths,
                lixi_transaction_type=self.lixi_transaction_type,
                lixi_version=self.lixi_version,
                output_name=None,
                output_folder=None,
                output_type=self.file_type,
            )
        else:

            basename = os.path.basename(output_path)
            dirname = os.path.dirname(output_path)

            if os.path.isdir(output_path):
                output_folder = output_path
                output_name = None
            elif os.path.isfile(output_path):
                output_folder = dirname
                output_name = (
                    str(basename)
                    .replace(".xsd", "")
                    .replace(".json", "")
                    .replace(".txt", "")
                )
            elif (
                ".json" in basename
                or ".xsd" in basename
                or ".xml" in basename
                or ".txt" in basename
            ):
                output_folder = dirname
                output_name = (
                    str(basename)
                    .replace(".xsd", "")
                    .replace(".json", "")
                    .replace(".txt", "")
                )
            else:
                output_folder = None
                output_name = None

            return _LIXI.getInstance().generate_custom_schema(
                instructions=restriction_paths,
                lixi_transaction_type=self.lixi_transaction_type,
                lixi_version=self.lixi_version,
                output_name=output_name,
                output_folder=output_folder,
                output_type=self.file_type,
            )

    def transform_message(self, xslt_folder_path, to_version=None, return_instance=False):
        """Transforms a LIXI Message instance to an older/newer version, based on the schema version specified, of the LIXI Message instance.
    
        Args:
            xslt_folder_path (:obj:`bool`, required): A path to a folder containing LIXI transformation (.xslt) files. 
            to_version (:obj:`str`, optional): The version of the LIXI schema to be transformed to. Should be in the format of '2.6.24'. Defaults to the latest version if not provided.
            return_instance (:obj:`bool`, optional): Indicates if a LIXI message instance needs to be created from the result and returned. 
    
        Returns:
            Transformed LIXI Message instance ONLY if return_instance is set to True.
    
        Raises:
            LIXIResouceNotFoundError: If the schema is not found at the schema path.
            LIXIInvalidSyntax: If the schema file is not well formed.
        """
        
        # Incase this is a json message, functionality works with xml only
        if self.xml_package == None:
            
            # Schema to be used as per message to derive paths from
            schema = _LIXI.getInstance().fetch_xml_schema(
                lixi_transaction_type=self.lixi_transaction_type,
                lixi_version=self.lixi_version,
                custom_version=self.lixi_custom_version,
                return_annotated = False 
            )
            schema = _etree.fromstring(schema)            
            
            self.xml_package = _xml_to_json.to_xml(self.json_package, schema)

        from_version_str = self.lixi_version
        transaction_type = self.lixi_transaction_type

        if to_version == None:
            
            if _LIXI.getInstance().schema_folder_path == None:
                raise LIXIResouceNotFoundError(
                    "Please specify the transformation version or please set the schema folder with lixi.set_schema_folder() .\n"
                    + str(error_message)
                )            
            
            to_version_str = str(
                _LIXI.getInstance().get_schema_latest_version(transaction_type)
            ).replace("_", ".")
        else:
            to_version_str = to_version

        result, all_warnings = _xslt_transform.transform_xslt(
            self.xml_package, xslt_folder_path, transaction_type, from_version_str, to_version_str
        )

        if return_instance == True:
            return Message(
                _etree.tostring(result, pretty_print=True).decode("utf-8"), None, "xml",
            )
        else:
            self.xml_package = result

    def get_transform_warnings(self, xslt_folder_path, to_version=None, output_path=None):
        """Gets the warnings when transforming a LIXI Message instance. Contains a record of data lost in the Message.
    
        Args:
            xslt_folder_path (:obj:`bool`, required): A path to a folder containing LIXI transformation (.xslt) files. 
            to_version (:obj:`str`, required): The version of the LIXI schema to be transformed to. Should be in the format of '2.6.24'. Defaults to the latest version if not provided.
            output_path (:obj:`str`, optional): Path to write the schema paths list to. Defaults to None.
    
        Returns:
            A string of xsl transform warnings.
    
        Raises:
            LIXIResouceNotFoundError: If the schema is not found at the schema path.
            LIXIInvalidSyntax: If the schema file is not well formed.
        """

        # Incase this is a json message, functionality works with xml only
        if self.xml_package == None:
            # Schema to be used as per message to derive paths from
            schema = _LIXI.getInstance().fetch_xml_schema(
                lixi_transaction_type=self.lixi_transaction_type,
                lixi_version=self.lixi_version,
                custom_version=self.lixi_custom_version,
                return_annotated = False
            )
            schema = _etree.fromstring(schema)            
            
            self.xml_package = _xml_to_json.to_xml(self.json_package, schema)

        from_version_str = self.lixi_version
        transaction_type = self.lixi_transaction_type

        if to_version == None:
            if _LIXI.getInstance().schema_folder_path == None:
                raise LIXIResouceNotFoundError(
                    "Please specify the transformation version or please set the schema folder with lixi.set_schema_folder() .\n"
                    + str(error_message)
                )            
            
            to_version_str = str(
                _LIXI.getInstance().get_schema_latest_version(transaction_type)
            ).replace("_", ".")
        else:
            to_version_str = to_version

        result, all_warnings = _xslt_transform.transform_xslt(
            self.xml_package, xslt_folder_path, transaction_type, from_version_str, to_version_str
        )

        if output_path == None:
            return all_warnings
        else:
            self.__write__(
                "transform_warnings_output.xml",
                output_path,
                all_warnings,
                "Message transform warnings failed. Can not store the schema at the specified folder.",
            )
            return all_warnings

    def validate_schematron(
        self, schematron_schema_text=None, schematron_schema_path=None
    ):
        """Validates this LIXI Message object against a Schematron file containing rules and assertions.
    
        Args:
            schematron_schema_text (:obj:`str`, required): Schematron rules schema provided as a text.
            schematron_schema_path (:obj:`str`, required): Schematron rules schema provided as a path.
    
        Result:
            result (:obj:`bool`): Indicates if validation was successfull.
            message (:obj:`str`): Validation message.
            error_log (:obj:`list`): Validation error object as a Python List.
    
        Raises:
            LIXIResouceNotFoundError: If the schema is not found at the schema path.
            LIXIInvalidSyntax: If the schema file is not well formed.
        """

        # Incase this is a json message, functionality works with xml only
        if self.xml_package == None:
            # Schema to be used as per message to derive paths from
            schema = _LIXI.getInstance().fetch_xml_schema(
                lixi_transaction_type=self.lixi_transaction_type,
                lixi_version=self.lixi_version,
                custom_version=self.lixi_custom_version,
                return_annotated = False
            )
            schema = _etree.fromstring(schema)            
            
            self.xml_package = _xml_to_json.to_xml(self.json_package, schema)

        if schematron_schema_path != None:
            if os.path.exists(schematron_schema_path) == True:
                f = io.open(schematron_schema_path, mode="r", encoding="utf-8")
                schematron_schema_text = f.read()
                f.close()
            else:
                raise LIXIResouceNotFoundError(
                    "Message schematron validation failed. Schematron rules schema file not found at the specified path."
                )

        try:
            parser = _etree.XMLParser(remove_blank_text=True)
            schematron_schema_etree = _etree.fromstring(schematron_schema_text, parser)
        except Exception as e:
            raise LIXIInvalidSyntax(
                "Message schematron validation failed. The schematron schema is not well-formed.\n\nError Message:\n    "
                + str(e)
            )

        result, message, error_log = _schematron_functions.validate(
            self.xml_package, schematron_schema_etree
        )

        return result, message, error_log

    def validate(self, schema=None, schema_path=None):
        """Validates this LIXI Message object against the corresponding LIXI schema.

        Args:
            schema (:obj:`str`, required): LIXI schema provided as a string. Defaults to None.  
            schema_path (:obj:`str`, required): LIXI schema provided as a path. Defaults to None.

        Result:
            result (:obj:`bool`): Indicates if validation was successfull.
            message (:obj:`str`): Validation message.
            error_log (:obj:`list`): Validation errors in list.

        Raises:
            LIXIValidationError: Validation errors for the xml file.
        """
        
        class Error_Log:
            """Represents a Error when validating with a schematron file.
                
            """
            def __init__(self, message, test, message_line, message_location):
                """Creates a LIXI message XML.
            
                Args:
                    message (:obj:`str`): Error string of the message.
                    test (:obj:`str`) : Test that caused the Error.
                    message_line (:obj:`str`): Line number in the input message.
                    message_location (:obj:`str`): Location xpath in the input message.
        
                """
        
                self.message = str.strip(message)
                self.test = test
                self.message_line = str(message_line)
                self.message_location = message_location        
        
        if schema!=None:
            schema_str = _LIXI.getInstance().fetch_xml_schema(
                schema_string=schema
            )
            schema_xsd = _etree.fromstring(schema_str)
            
        elif schema_path!=None:
            schema_str = _LIXI.getInstance().fetch_xml_schema(
                schema_path=schema_path,
            )
            schema_xsd = _etree.fromstring(schema_str)
        elif self.lixi_custom_version!=None:            
            schema_str = _LIXI.getInstance().fetch_xml_schema(
                custom_version=self.lixi_custom_version
            )
            schema_xsd = _etree.fromstring(schema_str)
        else:
            schema_str = _LIXI.getInstance().fetch_xml_schema(
                lixi_transaction_type=self.lixi_transaction_type,
                lixi_version=self.lixi_version
            ) 
            schema_xsd = _etree.fromstring(schema_str)

        validated_with = None
        if self.lixi_custom_version != None:
            schema_custom_version = schema_xsd.xpath(
                './xs:element[@name="Package"]/xs:complexType/xs:sequence/xs:element[@name="SchemaVersion"]/xs:complexType/xs:attribute[@name="LIXICustomVersion"]',
                namespaces=ns,
            )[0]
            if "fixed" not in schema_custom_version.attrib:
                raise LIXIInvalidSyntax(
                    "Version of the custom schema used is either incorrectly referenced in the message or does not specify the LIXICustomVersion. Please generate a new custom schema with lixi.get_custom_schema(). "
                )
            schema_custom_version = schema_custom_version.attrib["fixed"]

            validated_with = "LIXI Schema '" + schema_custom_version + "'."
            
        elif schema_path!=None or schema!=None:
            
            try:
                tr_type = schema_xsd.xpath(
                    './xs:element[@name="Package"]/xs:complexType/xs:sequence/xs:element[@name="SchemaVersion"]/xs:complexType/xs:attribute[@name="LIXITransactionType"]',
                    namespaces=ns,
                )[0]
                tr_type = tr_type.attrib["fixed"]
                
                i_version = schema_xsd.xpath(
                    './xs:element[@name="Package"]/xs:complexType/xs:sequence/xs:element[@name="SchemaVersion"]/xs:complexType/xs:attribute[@name="LIXIVersion"]',
                    namespaces=ns,
                )[0]
                i_version = i_version.attrib["fixed"]
                
                validated_with = (
                    "provided LIXI Schema '"
                    + tr_type
                    + " "
                    + i_version
                    + "'."
                )
            except:
                raise LIXIInvalidSyntax(
                    "Schema read failed. Invalid Schema. Missing LIXITransactionType or LIXIVersion. "
                )            
            
            schema_custom_version = schema_xsd.xpath(
                './xs:element[@name="Package"]/xs:complexType/xs:sequence/xs:element[@name="SchemaVersion"]/xs:complexType/xs:attribute[@name="LIXICustomVersion"]',
                namespaces=ns,
            )
            if len(schema_custom_version)>0:
                if "fixed" in schema_custom_version[0].attrib:
                    schema_custom_version = schema_custom_version.attrib["fixed"]       
                    validated_with = "provided LIXI Custom Schema '" + schema_custom_version + "'."
                                    
                
        else:
            validated_with = (
                "LIXI Schema '"
                + self.lixi_transaction_type
                + " "
                + self.lixi_version
                + "'."
            )

        if self.file_type == "xml":
            schema_validation = _etree.XMLSchema(schema_xsd)
            result = schema_validation.validate(self.xml_package)

            if result:
                self.is_valid = True
                self.validation_message = (
                    "Message is valid, validated with " + validated_with
                )
                return True, "Message is valid, validated with " + validated_with, []
            else:
                
                list_of_error = []
                
                error = (
                    "Message read failed. Message is not valid when validated with "
                    + validated_with
                    + "\n"
                )
                error = error + "\nInvalid Assertions:"
                for scerror in schema_validation.error_log:
                    error = (
                        error
                        + "\n    Error On Line "
                        + str(scerror.line)
                        + ": "
                        + scerror.message
                    )
                    list_of_error.append(Error_Log(scerror.message, 'Validation', scerror.line, scerror.path))

                self.is_valid = False
                self.validation_message = error
                return False, error, list_of_error

        elif self.file_type == "json":

            # XML provides a better check especially around dates and checking cross references.
            # Methods for json validation which are close to xml validation exist but up to user's discretion.
            if self.xml_package == None:
                self.xml_package = _xml_to_json.to_xml(self.json_package, schema_xsd)

            schema_validation = _etree.XMLSchema(schema_xsd)
            result = schema_validation.validate(self.xml_package)

            if result:
                self.is_valid = True
                self.validation_message = (
                    "Message file is valid, validated with " + validated_with
                )
                return True, "Message file is valid, validated with " + validated_with, []
            else:
                xml_errors = ""
                list_of_error = []

                for scerror in schema_validation.error_log:
                    xml_errors = (
                        xml_errors
                        + "    Error On Line "
                        + "JSONX"
                        + ": "
                        + scerror.message
                        + "\n"
                    )
                    list_of_error.append(Error_Log(scerror.message, 'Validation', scerror.line, scerror.path))

                errors = str(xml_errors).split("\n")

                json_msg = self.json_string.split("\n")

                json_error_string = (
                    "Message read failed. Message is not valid when validated with "
                    + validated_with
                    + "\n"
                )
                json_error_string = json_error_string + "\nInvalid Assertions:"

                for error in errors:

                    element = re.search(r"Element '([a-zA-Z. ]*)'", error)
                    attribute = re.search(r"attribute '([a-zA-Z. ]*)'", error)

                    if attribute != None:
                        attribute_check = attribute.group(1)
                    else:
                        attribute_check = None

                    if element != None:
                        element_check = element.group(1)
                    else:
                        element_check = None

                    line_no = 1
                    for line in json_msg:
                        if attribute_check != None:
                            if attribute_check in line:
                                json_error_string += "\n" + error.replace(
                                    "JSONX", str(line_no)
                                )
                                break

                        elif element_check != None:
                            if element_check in line:
                                json_error_string += "\n" + error.replace(
                                    "JSONX", str(line_no)
                                )
                                break

                        line_no = line_no + 1

                self.is_valid = False
                self.validation_message = json_error_string
                return False, json_error_string, list_of_error

    def to_string(self):
        """Returns a string representation of the LIXI message.

        Returns:
            A pretty printed LIXI Message string.
        """

        if self.file_type == "xml":
            return str(
                _etree.tostring(self.xml_package, pretty_print=True).decode("utf-8")
            ).strip()
        elif self.file_type == "json":
            return str(
                _json.dumps(
                    self.json_package, sort_keys=True, indent=4, ensure_ascii=False
                )
            ).strip()

    def pretty_print(self):
        """Serialises to a string and prints the LIXI Message object.

        Result:
            Pretty prints a LIXI Message.
        """
        print(self.to_string())

    def save(self, output_path=None):
        """Serialises and saves a LIXI Message object to a file using the path provided.

        Args:
            output_path (:obj:`str`, optional): Path to write the schema paths list to. Defaults to None.
            
        Raises:
            LIXIResouceNotFoundError: If the path provided is not valid.    
        """

        string = self.to_string()

        if output_path == None:
            if self.file_type == "xml":
                output_path = "message_output.xml"
            elif self.file_type == "json":
                output_path = "message_output.json"

        self.__write__(
            "message_output.txt",
            output_path,
            string,
            "Message save failed. Can not store the schema at the specified folder.",
        )



import os, re, filecmp, sys, datetime, html, inspect, lxml, glob, zipfile, io, xml.dom.minidom as dome
from operator import attrgetter
from shutil import copyfile
from lxml import etree
from lxml import builder

ns = {
    "xs": "http://www.w3.org/2001/XMLSchema",
    "lx": "lixi.org.au/schema/appinfo_elements",
    "li": "lixi.org.au/schema/appinfo_instructions",
}


def apply_xslt(xmlFile, xsltFile):
    """ Applys a xslt rule file to a XML message.
    
    Args:
        xmlFile (:obj:`lxml.etree`, required): A XML Message.
        xsltFile (:obj:`lxml.etree`, required): A XSLT transform rules file.
    
    Returns:
        A string of result XML.
    
    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    xslt = etree.parse(xsltFile)
    transform = etree.XSLT(xslt)
    newdom = transform(xmlFile)
    outstr = etree.tostring(
        newdom, pretty_print=True, xml_declaration=True, encoding="UTF-8"
    )

    return outstr


def transform_xslt(sample_etree, xslt_folder_path, transaction_type, from_version_str, to_version_str):
    """ Transforms a LIXI Message instance to an older/newer version, based on the schema version specified, of the LIXI Message instance.
    
    Args:
        sample_etree (:obj:`str`, required): A LIXI Message provided as a string. Defaults to None.
        xslt_folder_path (:obj:`bool`, required): A path to a folder containing LIXI transformation (.xslt) files. 
        transaction_type (:obj:`bool`, required): The transaction type of a LIXI schema. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None.
        from_version_str (:obj:`str`, required): The version of the LIXI schema to be transformed to. Should be in the format of '2.6.24'. 
        to_version_str (:obj:`str`, required): The version of the LIXI schema to be transformed to. Should be in the format of '2.6.24'. 

    Returns:
        Transformed LIXI Message instance ONLY if return_instance is set to True.

    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    def sort_versions(tranforms):
        temp_list = []

        def num_keyfn(x):
            parts = x.split("_")
            return tuple(int(x) if x.isnumeric else x for x in parts)

        for num in sorted(tranforms, key=num_keyfn):
            temp_list.append(num)

        return temp_list

    # Determine if its a backwards or forwards transformation (assumption, version provided by user is of the form '#.#.#')
    to_version = to_version_str.split(".")
    from_version = from_version_str.split(".")

    if int(to_version[0]) == int(from_version[0]):
        if int(to_version[1]) == int(from_version[1]):
            if int(to_version[2]) == int(from_version[2]):
                return sample_etree, ""
            elif int(to_version[2]) > int(from_version[2]):
                direction = "forwards"
            elif int(to_version[2]) < int(from_version[2]):
                direction = "backwards"
        elif int(to_version[1]) > int(from_version[1]):
            direction = "forwards"
        elif int(to_version[1]) < int(from_version[1]):
            direction = "backwards"
    elif int(to_version[0]) > int(from_version[0]):
        direction = "forwards"
    elif int(to_version[0]) < int(from_version[0]):
        direction = "backwards"

    # Vars help order the final list correctly
    if direction == "forwards":
        greater_version_no = (
            "LIXI-" + transaction_type + "-" + to_version_str.replace(".", "_")
        )
        lesser_version_no = (
            "LIXI-" + transaction_type + "-" + from_version_str.replace(".", "_")
        )
    else:
        greater_version_no = (
            "LIXI-" + transaction_type + "-" + from_version_str.replace(".", "_")
        )
        lesser_version_no = (
            "LIXI-" + transaction_type + "-" + to_version_str.replace(".", "_")
        )

    # The complete version lists
    tranforms = []
    #for filename in glob.glob(
        #os.path.join(pkg_resources.resource_filename(__name__, "xslts"), "*.xsl")
    #):
    for filename in glob.glob(
        os.path.join(xslt_folder_path, "*.xsl")
    ):        
        if "warnings" not in filename and transaction_type in filename:
            filename = re.search(
                r"LIXI-([a-zA-Z0-9_-]*).xsl", filename.replace("LIXI", "", 1)
            ).group(1)
            filename = filename.replace(transaction_type + "-", "")
            if filename not in tranforms:
                tranforms.append(filename)
    tranforms = sort_versions(tranforms)
    tranforms_versions = [
        ("LIXI-" + transaction_type + "-" + x.replace(".", "_")) for x in tranforms
    ]

    # Create a list of only required version/transaction strings and in the correct order.
    count_forwards = []
    count_backwards = []
    tranforms_list = []
    check = False

    for version in tranforms_versions:
        if version == lesser_version_no or check == True:
            count_forwards.append(version)
            check = True

    check = False
    count_forwards = list(reversed(count_forwards))
    for version in count_forwards:
        if version == greater_version_no or check == True:
            count_backwards.append(version)
            check = True

    if direction == "forwards":
        tranforms_list = list(reversed(count_backwards))
    else:
        tranforms_list = count_backwards

    tranforms_forwards, tranforms_backwards = (
        count_backwards,
        list(reversed(count_backwards)),
    )

    if direction == "forwards":
        tranforms_forwards, tranforms_backwards = (
            tranforms_backwards,
            tranforms_forwards,
        )

    # The actual transformation of the files from one version to all prev versions
    step = 0

    temp_prev_step_xml = sample_etree
    temp_current_step_xml = None
    temp_step_warnings_xml = etree.Element(
        "{http://www.w3.org/2001/XMLSchema-instance}warnings"
    )

    for first, second in zip(tranforms_list, tranforms_list[1:]):

        #archive = zipfile.ZipFile(
            #os.path.join(
                #pkg_resources.resource_filename(__name__, "xslts"),
                #first + "-" + second + ".zip",
            #),
            #"r",
        #)
        #stylesheet = io.BytesIO(archive.read(first + "-" + second + ".xsl"))
        stylesheet = os.path.join(xslt_folder_path,first + "-" + second + ".xsl")
        
        #archive = zipfile.ZipFile(
            #os.path.join(
                #pkg_resources.resource_filename(__name__, "xslts"),
                #first + "-" + second + "-warnings.zip",
            #),
            #"r",
        #)
        #warningsheet = io.BytesIO(archive.read(first + "-" + second + "-warnings.xsl"))
        warningsheet = os.path.join(xslt_folder_path, first + "-" + second + "-warnings.xsl")
        
        # schemafile   = os.path.join(xsd_folder_path, second + "-Annotated.xsd")

        # print(stylesheet + "\t\t" + "-step-" + str(step + 1), end="")

        # do the XSL transformation
        temp_current_step_xml = etree.fromstring(
            apply_xslt(temp_prev_step_xml, stylesheet)
        )

        # step
        comment = etree.Comment(" Step " + str(step) + " ")
        temp_step_warnings_xml.append(comment)

        # create warnings for this transformation
        temp_current_step_warnings_xml = etree.fromstring(
            apply_xslt(temp_prev_step_xml, warningsheet)
        )
        for child in temp_current_step_warnings_xml:
            temp_step_warnings_xml.append(child)

        # Validate the resulting XML against the correct schema
        # schema_etree = etree.parse(schemafile)
        # schema = etree.XMLSchema(schema_etree)
        # valid = schema.validate(temp_current_step_xml)
        # valid = True
        # if valid:
        # print("\t...valid")

        # else:
        # print("\n")
        # for error in schema.error_log:
        # print("ERROR ON LINE "+str(error.line)+": "+str(error.message.encode("utf-8"))+".")
        # raise Exception("ERROR: tranformed message is not valid, "+str(step + 1)+".")

        temp_prev_step_xml = temp_current_step_xml
        step = step + 1

    xml_string = etree.tostring(temp_step_warnings_xml).decode()
    xml_string = xml_string.strip().replace(">\n", ">").replace("\n<", "<")
    reparsed = dome.parseString(xml_string)
    all_warnings = reparsed.toprettyxml(indent="  ")

    return temp_current_step_xml, all_warnings



import importlib.resources
import os as _os
import io as _io
import json as _json
#import boto3 as _boto3
import glob as _glob
from lxml import etree as _etree
import time, re
from itertools import islice


ns = {
    "xs": "http://www.w3.org/2001/XMLSchema",
    "lx": "lixi.org.au/schema/appinfo_elements",
    "li": "lixi.org.au/schema/appinfo_instructions",
}

if __name__ == "_LIXI":
    import _path_functions, _jsonschema_functions, _customise_schema, _xslt_transform, _schema_documentation, _explode_schema
else:
    from lixi import (
        _path_functions,
        _jsonschema_functions,
        _customise_schema,
        _xslt_transform, 
        _schema_documentation,
        _explode_schema
    )


class LIXI(object):
    """Represents a LIXI Instance.
        
    The singleton class exists as a wrapper for all message manipulation functions the library is to provide.
    A LIXI schema is required because it is the underlying assumption that all LIXI members would 
    have a LIXI schema.
    """
    __instance = None

    @staticmethod
    def getInstance():
        """ Static access method. """
        if LIXI.__instance == None:
            LIXI()
        return LIXI.__instance

    def __init__(self):
        """ Virtually private constructor. """

        if LIXI.__instance != None:
            print("This should not have run")
        else:

            self.schema_folder_path = (
                None  # _pkg_resources.resource_filename(__name__, 'schema/schema')
            )
            self.schema_index_dict = {}

            self.__S3_access = None
            self.__S3_secret = None

            self.current_transaction_type = None

            self.__schemas = {}
            self.__latest_versions = {}
            #self.__set_schema_latest_version__()
            
            self._lambda_check = False # used to clock all config files updates for lambda

            LIXI.__instance = self

    ####################
    ## HELPER FUNCTIONS
    ####################
    def __write__(self, default_name, output_path, data, error_string):
        """Helper function to write to a file.

        Args:
            default_name (:obj:`str`, required): A default name in case the provided out path is a folder path and not a complete file path.
            output_path (:obj:`str`, required): The output path of the file to be written.
            data (:obj:`str`, required): Data to be written to the  file.
            error_string (:obj:`str`, required): A context specific error string in case write fails.

        Raises:
            LIXIResouceNotFoundError: If output path provided does not exist.
        """

        if _os.path.isdir(output_path):
            output_path = _os.path.join(output_path, default_name)

        try:
            with _io.open(output_path, "w+", encoding="utf-8") as out_file:
                out_file.write(data)
        except Exception as e:
            raise LIXIResouceNotFoundError(error_string)

    def __set_schema_latest_version__(self):
        """Helper function to set the latest version of the schema per transaction type. Relies on stored transformation files to obtain latest transaction version.

        Raises:
            LIXIResouceNotFoundError: If output path provided does not exist.
        """

        # Relies on stored transformation files to obtain latest transaction version. So these need to be updated every time there is a new version. Runs at program start.
        for filename in _glob.glob(
            # _os.path.join(_pkg_resources.resource_filename(__name__, "xslts"), "*.zip")
            _os.path.join(importlib.resources.files(__name__).parent, "xslts", "*.zip")
        ):
            if "warnings" not in filename:
                filename = re.search(
                    r"LIXI-([a-zA-Z0-9_-]*).zip", filename.replace("LIXI", "", 1)
                ).group(1)

                split = filename.split("-")

                transaction_type = split[0]
                digit_1, digit_2, digit_3 = split[1].split("_")

                if transaction_type in self.__latest_versions:
                    old_version = self.__latest_versions[transaction_type]
                    olddigit_1, olddigit_2, olddigit_3 = old_version.split("_")

                    if int(digit_1) > int(olddigit_1):
                        self.__latest_versions[transaction_type] = split[1]
                    elif int(digit_1) == int(olddigit_1):
                        if int(digit_2) > int(olddigit_2):
                            self.__latest_versions[transaction_type] = split[1]
                        elif int(digit_2) == int(olddigit_2):
                            if int(digit_3) > int(olddigit_3):
                                self.__latest_versions[transaction_type] = split[1]

                else:
                    self.__latest_versions[transaction_type] = split[1]
    
    def __get_folder_latest_version__(self, lixi_transaction_type, actual_latest_schema_version):
        """Helper function to get the latest version of the schema per transaction type in a schema folder IF the latest actual latest schema does not exist locally. Needs to have a schema folder already specified.

        Args:
            lixi_transaction_type (:obj:`str`, required): Transaction type of the LIXI schema.
            actual_latest_schema_version (:obj:`str`, required): Latest version of the transation schema.

        Returns:
            A schema version string with format (X_X_X).
        """
        
        # Check if config file exists and use it
        if _os.path.exists(self.schema_folder_path + "/config.json"):
            with _io.open(
                self.schema_folder_path + "/config.json", encoding="utf-8"
            ) as json_index:
                schema_index_dict = _json.load(json_index)           
        
        list_of_schema_keys = schema_index_dict.keys()
        
        # If latest version is in folder proceed to return that.
        if 'LIXI-'+lixi_transaction_type+"-"+str(actual_latest_schema_version)+'-Annotated' in list_of_schema_keys:
            return actual_latest_schema_version
        
        # If not then try to derive the folder latest number and return.
        else:
            list_of_relevant_schema_keys = [str(key).replace('LIXI-'+lixi_transaction_type+'-','').replace("-Annotated","") for key in list_of_schema_keys if lixi_transaction_type in key and 'Custom' not in key]
            
            if len(list_of_relevant_schema_keys) == 0:
                return actual_latest_schema_version
            
            folder_latest_schema_version = None
            
            for key in list_of_relevant_schema_keys:
                if folder_latest_schema_version != None:
                    digit_1, digit_2, digit_3 = key.split("_")
                    
                    old_version = folder_latest_schema_version
                    olddigit_1, olddigit_2, olddigit_3 = old_version.split("_")

                    if int(digit_1) > int(olddigit_1):
                        folder_latest_schema_version = key 
                    elif int(digit_1) == int(olddigit_1):
                        if int(digit_2) > int(olddigit_2):
                            folder_latest_schema_version = key 
                        elif int(digit_2) == int(olddigit_2):
                            if int(digit_3) > int(olddigit_3):
                                folder_latest_schema_version = key 

                else:
                    folder_latest_schema_version = key                 
            
            return folder_latest_schema_version    
    
    
    def __create_schema_index_from_folder__(self, reset=False, create_config=True):
        """Helper function to create a config file (record of local schemas) from a folder.

        Args:
            reset (:obj:`boolean`, optional): A flag to force re-creation of the config file. Defaults to None.
            create_config (:obj:`boolean`, optional): Variable to force creation of config file for a schema folder. Defaults to True. Only for internal source.

        Raises:
            LIXIResouceNotFoundError: If schema folder path provided does not exist.
            LIXIInvalidSyntax: Validation errors for the lixi version or transaction type.
        """

        # if reset == false proceed to read from a pre-built config file.
        if _os.path.exists(self.schema_folder_path + "/config.json") and reset == False or self._lambda_check == True:
            with _io.open(
                self.schema_folder_path + "/config.json", encoding="utf-8"
            ) as json_index:
                self.schema_index_dict = _json.load(json_index)
        elif create_config and self._lambda_check==False:
            # read from the folder path and create a config file.

            if _os.path.exists(self.schema_folder_path) == False:
                raise LIXIResouceNotFoundError(
                    "Schema read failed. Schema folder path specified does not exist."
                )

            files = _glob.glob(_os.path.join(self.schema_folder_path, "*.xsd"))

            for filename in files:

                filename = filename.replace("\\", "/")
                f = _io.open(filename, mode="r", encoding="utf-8")
                
                if 'LIXI-Master-Schema.xsd' not in filename:
                
                    is_schema = False
                    read_string = ''
                    
                    # Read the first twenty-five lines of a schema to determine its transaction and version from annotation. Max is 25 lines.
                    for line in islice(f, 25):
                        read_string = read_string + line
                    
                    if "lx:schemadetail" in read_string:
                        schemadetail = re.search(
                            r'<lx:schemadetail(.*?)>', read_string
                        )
                        
                        if schemadetail != None:
                            schemadetail = schemadetail.group(1)
                            schemadetail = _etree.fromstring("<INFO "+schemadetail+">")
                            
                            if schemadetail.get('type') != "Master":
                                if schemadetail.get('type') == "Custom":
                                    schema_custom_version = schemadetail.get('version')
                                    annotated = schemadetail.get('annotation')
                                    is_schema = True
                                elif schemadetail.get('type') == "Transaction":
                                    schema_custom_version = None
                                    schema_transaction_type = schemadetail.get('transaction')
                                    schema_version = schemadetail.get('version')
                                    annotated = schemadetail.get('annotation')
                                    is_schema = True
    
                    # 1) only take a annotated schema 2) create a different key name so that filename won't matter 3) also if schema is custom version store that name 4) else store it as per transaction - version
                    if is_schema == True:
                        if schema_custom_version != None:
                            self.schema_index_dict[schema_custom_version] = filename
                        else:
                            if annotated == "Full":
                                self.schema_index_dict[
                                    "LIXI-"
                                    + schema_transaction_type
                                    + "-"
                                    + schema_version.replace(".", "_")
                                    + "-Annotated"
                                ] = filename
                            else:
                                self.schema_index_dict[
                                    "LIXI-"
                                    + schema_transaction_type
                                    + "-"
                                    + schema_version.replace(".", "_")
                                ] = filename                                
                
                f.close()
            # config file is stored. create_config will need to be turned off when used with AWS where write permissions are turned off.
            with _io.open(self.schema_folder_path + "/config.json", "wb+") as outfile:
                s = _json.dumps(
                    self.schema_index_dict, sort_keys=True, indent=4, ensure_ascii=False
                ).encode("utf-8")
                outfile.write(s)

        else:
            raise LIXIInvalidSyntax(
                "Schema read failed. Schema config file could not be created."
            )

    def __parse_xml_schema__(self, xsd_as_text, file_type, schema_internal_source):
        """Helper function to parse a LIXI XML schema from a string object. Important to note, will always return LIXI XML Etree Object.

        Args:
            xsd_as_text (:obj:`str`, required): LIXI schema provided as a string.
            file_type (:obj:`str`, optional): The type of the input LIXI schema given. Defaults to 'xml'. DEPRECATED. But have kept it for later use if needed.
            schema_internal_source (:obj:`str`, optional): A description of the source of the schema. Used for more descriptive error messages, and debugging.

        Returns:
             (
                'schema_output': A LIXI XML Etree object,
                'schema_transaction_type': The transaction type of the LIXI schema,
                'schema_version': The version of the transaction LIXI schema,
                'schema_custom_version': The custom version of the LIXI schema,
                'is_annotated' : Variable to indicate if the schema is annotated or not.
             )

        Raises:
            LIXIInvalidSyntax: Schema can not be read or type is not supported.
            LIXIInvalidSyntax: Missing Transaction Type.
            LIXIInvalidSyntax: Missing LIXI Version.
        """

        # Tries to parse the string as xml or json (converts back to xml schema since only xml is processed)
        passed = False
        schema_output = None

        if passed == False:
            try:
                parser = _etree.XMLParser(remove_blank_text=True)
                schema_output = _etree.fromstring(xsd_as_text, parser)
                passed = True
            except Exception as e:
                passed = False

        if passed == False:
            try:
                self.json_package = _json.loads(xsd_as_text)
                schema_output = _jsonschema_functions.convert_to_xml_schema(xsd_as_text)
                passed = True
            except Exception as e:
                passed = False

        # Parsing failed now we need better error messages
        if passed == False:

            error_message = ""

            if "<xs:schema" in xsd_as_text:
                try:
                    doc = etree.parse(filename_xml)
                except Exception as e:
                    error_message = e
            elif '"$schema"' in xsd_as_text:  ## is json
                try:
                    _json.loads(str(xsd_as_text))
                except Exception as e:
                    error_message = e

            raise LIXIInvalidSyntax(
                "Schema read failed. Schema from source '"
                + str(schema_internal_source)
                + "' can not be read or type is not supported.\n\nError Message:\n    "
                + str(error_message)
            )

        # Parsing suceeded now we need to extract schema info
        SchemaVersion = schema_output.xpath(
            './xs:element/xs:complexType/xs:sequence/xs:element[@name="SchemaVersion"]',
            namespaces=ns,
        )[0]

        try:
            schema_transaction_type = SchemaVersion.xpath(
                './xs:complexType/xs:attribute[@name="LIXITransactionType"]',
                namespaces=ns,
            )[0]
            schema_transaction_type = schema_transaction_type.attrib["fixed"]
            self.current_transaction_type = schema_transaction_type
        except Exception as e:
            raise LIXIInvalidSyntax("Schema read failed. Missing Transaction Type.")

        try:
            schema_version = SchemaVersion.xpath(
                './xs:complexType/xs:attribute[@name="LIXIVersion"]', namespaces=ns
            )[0]
            schema_version = schema_version.attrib["fixed"]
        except Exception as e:
            raise LIXIInvalidSyntax("Schema read failed. Missing LIXI Version.")

        try:
            schema_custom_version = SchemaVersion.xpath(
                './xs:complexType/xs:attribute[@name="LIXICustomVersion"]',
                namespaces=ns,
            )[0]
            schema_custom_version = schema_custom_version.attrib["fixed"]
        except Exception as e:
            schema_custom_version = None
        
        try:
            annotated_val = schema_output.xpath(
                './xs:annotation/xs:appinfo/lx:schemadetail',
                namespaces=ns,
            )[0]
            annotated_val = annotated_val.attrib["annotation"]
            if annotated_val == "Minimal":
                is_annotated = False
            elif annotated_val == "Full":
                is_annotated = True
            
        except Exception as e:
            is_annotated = False        
        
        return (
            schema_output,
            schema_transaction_type,
            schema_version,
            schema_custom_version,
            is_annotated
        )

    def __read_path__(self, file_path, file_type="xml"):
        """Helper function to read a LIXI XML schema file from a path. Important to note, will always return LIXI XML Etree Object.

        Args:
            file_path (:obj:`str`, required): LIXI schema provided as a path.
            file_type (:obj:`str`, optional): The type of the input LIXI schema given. Defaults to 'xml'. DEPRECATED. But have kept it for later use if needed.

        Returns:
             (
                'schema_output': A LIXI XML Etree object,
                'schema_transaction_type': The transaction type of the LIXI schema,
                'schema_version': The version of the transaction LIXI schema,
                'schema_custom_version': The custom version of the LIXI schema,
                'is_annotated' : Variable to indicate if the schema is annotated or not.
             )

        Raises:
            LIXIResouceNotFoundError: Schema file not found at path.
        """

        # Read the file from path and display an error if not found.
        cwd = _os.getcwd()
        file_path = _os.path.join(cwd, file_path)

        if _os.path.exists(file_path) == True:
            f = _io.open(file_path, mode="r", encoding="utf-8")
            schema_string = f.read()
            f.close()
            (
                schema_output,
                schema_transaction_type,
                schema_version,
                schema_custom_version,
                is_annotated
            ) = self.__parse_xml_schema__(schema_string, file_type, file_path)

            return (
                schema_output,
                schema_transaction_type,
                schema_version,
                schema_custom_version,
                is_annotated
            )
        else:
            raise LIXIResouceNotFoundError(
                "Schema read failed. Schema file not found at the path "
                + str(file_path)
                + "."
            )

    def __load_schema__(
        self,
        lixi_transaction_type=None,
        lixi_version=None,
        file_type="xml",
        custom_version=None,
        schema_string=None,
        schema_path=None,
        create_config=True,
        return_annotated = False
    ):
        """Helper function to load a LIXI schema from multiple sources. Important to note, will always return LIXI XML Etree Object.

        Args:
            lixi_transaction_type (:obj:`str`, optional): The transaction type of the LIXI schema to be fetched. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None.
            lixi_version (:obj:`str`, optional): The version of the LIXI schema to be fetched. Should be in the format of '2.6.24'. Defaults to None. Defaults to the latest version if only lixi_transaction_type is provided.
            file_type (:obj:`str`, optional): The type of the input LIXI schema given. Defaults to 'xml'. DEPRECATED. But have kept it for later use if needed.
            custom_version (:obj:`str`, optional): The version of the LIXI custom schema to be fetched. Usually a complete file name of the custom schema. Defaults to None.
            schema_string (:obj:`str`, optional): LIXI schema provided as a string. Defaults to None.
            schema_path (:obj:`str`, optional): LIXI schema provided as a path. Defaults to None.
            create_config (:obj:`boolean`, optional): Variable to force creation of config file for a schema folder. Defaults to True. Only for internal source.
            return_annotated (:obj:`boolean`, optional): Variable to indicated if the returned schema should be annotated or not.

        Returns:
            A LIXI schema in the python Etree format.

        Raises:
            LIXIResouceNotFoundError: If no source of LIXI schema is provided.
            LIXIInvalidSyntax: Validation errors for the lixi version or transaction type.
            LIXIResouceNotFoundError: If path provided does not exist.
            LIXIValidationError: If path provided does not exist.
        """

        # Before loading the schema (source unknown at this point) make sure schema folder is read and a config file (record of local schemas) created.
        if (
            len(self.schema_index_dict) <= 0
            and self.schema_folder_path != None
            and create_config
        ):
            self.__create_schema_index_from_folder__(
                reset=False, create_config=create_config
            )

        # Read the schema into a Etree from the provided source.
        # Reading schema when it is passed as a string.
        if schema_string != None:
            (
                schema_output,
                schema_transaction_type,
                schema_version,
                schema_custom_version,
                is_annotated
            ) = self.__parse_xml_schema__(schema_string, file_type, "text")

            # Store it for later use if needed
            if schema_custom_version != None:
                self.__schemas[schema_custom_version] = schema_output
                return schema_output
            else:
                if is_annotated == True:
                    self.__schemas[
                        "LIXI-"
                        + schema_transaction_type
                        + "-"
                        + schema_version.replace(".", "_")
                        + "-Annotated"
                    ] = schema_output
                else:
                    self.__schemas[
                        "LIXI-"
                        + schema_transaction_type
                        + "-"
                        + schema_version.replace(".", "_")
                    ] = schema_output                     

                return schema_output

        # Reading schema when it is passed as a path.
        elif schema_path != None:
            (
                schema_output,
                schema_transaction_type,
                schema_version,
                schema_custom_version,
                is_annotated
            ) = self.__read_path__(schema_path, file_type)

            # Store it for later use if needed
            if schema_custom_version != None:
                self.__schemas[schema_custom_version] = schema_output
                return schema_output
            else:
                if is_annotated == True:
                    self.__schemas[
                        "LIXI-"
                        + schema_transaction_type
                        + "-"
                        + schema_version.replace(".", "_")
                        + "-Annotated"
                    ] = schema_output
                else:
                    self.__schemas[
                        "LIXI-"
                        + schema_transaction_type
                        + "-"
                        + schema_version.replace(".", "_")
                    ] = schema_output                     

                return schema_output

        # Reading a custom schema version. HAS to be read before reading schema from transaction and version because all custom schemas have these two fields populated so there is a risk of fetching the transaction schema when a custom schema is needed.
        elif custom_version != None:

            # If custom schema stored in local variable use that
            if custom_version in self.__schemas:
                return self.__schemas[custom_version]

            # Else look for it locally
            if self.schema_folder_path != None:

                # Update folder just in case custom schema was recently generated and stored
                if custom_version not in self.schema_index_dict:
                    self.__create_schema_index_from_folder__(
                        reset=True, create_config=create_config
                    )

                if custom_version in self.schema_index_dict:
                    filepath = self.schema_index_dict[custom_version]

                    (
                        schema_output,
                        schema_transaction_type,
                        schema_version,
                        schema_custom_version,
                        is_annotated
                    ) = self.__read_path__(filepath, file_type)

                    self.__schemas[schema_custom_version] = schema_output

                    return self.__schemas[schema_custom_version]
                else:
                    raise LIXIResouceNotFoundError(
                        "Schema read failed. Custom schema file "+ custom_version+ " not found in the schema folder."
                    )
            else:
                raise LIXIResouceNotFoundError(
                    "Schema read failed. Can not read custom schema file '"
                    + custom_version
                    + "'. Please set a schema folder with lixi.set_schema_folder()."
                )

        # Reading a schema with lixi transaction schema and version.
        elif lixi_transaction_type != None and lixi_version != None:
            
            schema_name_annotated = (
                "LIXI-"
                + lixi_transaction_type
                + "-"
                + lixi_version.replace(".", "_")
                + "-Annotated"
            )
            schema_name_unannotated = (
                "LIXI-"
                + lixi_transaction_type
                + "-"
                + lixi_version.replace(".", "_")
            )                
            self.current_transaction_type = lixi_transaction_type

            # Look for schema in stored variable
            
            if schema_name_annotated in self.__schemas and schema_name_unannotated in self.__schemas:
                if return_annotated == True:
                    return self.__schemas[schema_name_annotated]
                elif return_annotated == False:
                    return self.__schemas[schema_name_unannotated]
            elif schema_name_annotated in self.__schemas and schema_name_unannotated not in self.__schemas:
                if return_annotated == True:
                    return self.__schemas[schema_name_annotated]
                elif return_annotated == False:
                    return self.__schemas[schema_name_annotated]                    
            elif schema_name_annotated not in self.__schemas and schema_name_unannotated in self.__schemas:
                if return_annotated == True:
                    schema_name = schema_name_annotated
                elif return_annotated == False:
                    return self.__schemas[schema_name_unannotated]            
            elif schema_name_annotated not in self.__schemas and schema_name_unannotated not in self.__schemas:
                if return_annotated == True:
                    schema_name = schema_name_annotated
                elif return_annotated == False:
                    schema_name = schema_name_unannotated           
                    

            if self.schema_folder_path != None:

                
                if schema_name not in self.schema_index_dict:
                    # Update folder just in case custom schema was recently generated and stored
                    if create_config:
                        self.__create_schema_index_from_folder__(
                            reset=True, create_config=create_config
                        )

                if schema_name in self.schema_index_dict and create_config:
                    filepath = self.schema_index_dict[schema_name]
                    (
                        schema_output,
                        schema_transaction_type,
                        schema_version,
                        schema_custom_version,
                        is_annotated
                    ) = self.__read_path__(filepath, file_type)

                    self.__schemas[schema_name] = schema_output

                    return self.__schemas[schema_name]
                
                elif schema_name_annotated in self.schema_index_dict and create_config:
                    filepath = self.schema_index_dict[schema_name_annotated]
                    (
                        schema_output,
                        schema_transaction_type,
                        schema_version,
                        schema_custom_version,
                        is_annotated
                    ) = self.__read_path__(filepath, file_type)

                    self.__schemas[schema_name_annotated] = schema_output

                    return schema_output                
                else:
                    anno = ''
                    if return_annotated == True:
                        anno = 'Annotated'
                    
                    raise LIXIResouceNotFoundError(
                        "Schema read failed. Schema file "+ lixi_transaction_type+" "+ lixi_version+" "+anno+" not found in the schema folder."
                    )

            # Download the schema if credentials are provided
            #if self.__S3_access != None and self.__S3_secret != None:
                #session = _boto3.Session(
                    #aws_access_key_id=self.__S3_access,
                    #aws_secret_access_key=self.__S3_secret,
                #)

                #if file_type == "xml":
                    #schema_name2 = schema_name + ".xsd"

                #s3 = session.resource("s3")
                #obj = s3.Object("lixi-schema", schema_name2)
                #string_data = obj.get()["Body"].read().decode("utf-8")

                #(
                    #schema_output,
                    #schema_transaction_type,
                    #schema_version,
                    #schema_custom_version,
                    #is_annotated
                #) = self.__parse_xml_schema__(string_data, file_type, "S3")

                #self.__schemas[schema_name] = schema_output

                #return self.__schemas[schema_name]

            if (
                self.__S3_access == None
                and self.__S3_secret == None
                and self.schema_folder_path == None
            ):
                raise LIXIResouceNotFoundError(
                    "Schema read failed. Can not read schema file '"
                    + lixi_transaction_type
                    + " "
                    + lixi_version
                    + " "
                    + 'Annotated' if return_annotated == True else ''
                    + "'. Please set a schema folder with lixi.set_schema_folder()."
                )

        else:
            raise LIXIResouceNotFoundError(
                "Schema read failed. Please specify a LIXI Schema source (String, Path to LIXI schema, Path to a folder that contains a LIXI schema or LIXI schema S3 credentials)."
            )

    ###################
    ## ENTRY POINTS
    ###################
    #def set_credentials(self, access, secret):
        #"""Sets the secret and access keys for a LIXI member, which would enable a schema to be fetched from the LIXI online repository. 
        #Both access key and secret key can be requested from LIXI admin. These are unique for an organization/member. 
        
        #Args:
            #access (:obj:`str`, required): A LIXI member's assigned access key.
            #secret (:obj:`str`, required): A LIXI member's assigned secret key.
        
        #Raises:
            #LIXIInvalidSyntax: If 'access' is not specified or is not str.
            #LIXIInvalidSyntax: If 'secret' is not specified or is not str.
        #"""

        #if access != None and secret != None:

            #if type(access).__name__ == "str" and type(secret).__name__ == "str":
                #self.__S3_access = access
                #self.__S3_secret = secret
            #else:
                #raise LIXIInvalidSyntax("Parameter provided is not a string.")
        #else:
            #raise LIXIInvalidSyntax("Required parameter not specified.")

    def set_schema_path(self, schemapath, create_config=True, force_update=False, lambda_check=False):
        """Sets a directory path to fetch LIXI schemas from. Used before doing any operations that require a schema. Allows schema read from folder and a re-read of folder in case a file was added on runtime.  
    
        Args:
            folder_path (:obj:`str`, required): A path to a folder containing LIXI schema files.
            create_config (:obj:`bool`, required): A flag to indicate whether a folder read is required.
            force_update (:obj:`bool`, required): A flag to force folder read.
        
        Raises:
            LIXIInvalidSyntax: If path is not specified.
            LIXIResouceNotFoundError: If path provided does not exist.
        """

        if schemapath != None:
            if _os.path.exists(schemapath):
                self.schema_folder_path = str(schemapath)
                self.__create_schema_index_from_folder__(reset = force_update, create_config = create_config)
                self._lambda_check = lambda_check
            else:
                raise LIXIResouceNotFoundError("Schema folder path is incorrect.")
        else:
            raise LIXIResouceNotFoundError("Schema folder path not specified.")

    def get_schema_latest_version(self, lixi_transaction_type):
        """Gets the latest version number of a LIXI transation schema released, or the latest version number in the folder if a schema folder path has been previously provided.

        Args:
            lixi_transaction_type (:obj:`str`, required): Transaction type of the LIXI schema.

        Returns:
            A schema version string with format (X_X_X).

        Raises:
            LIXIInvalidSyntax: Invalid transaction type provided.
        """
        latest_version_of_schema = None 
        
        #if lixi_transaction_type == "ACC":
            #latest_version_of_schema = self.__latest_versions["ACC"]
        #elif lixi_transaction_type == "CAL":
            #latest_version_of_schema = self.__latest_versions["CAL"]
        #elif lixi_transaction_type == "CDA":
            #latest_version_of_schema = self.__latest_versions["CDA"]
        #elif lixi_transaction_type == "CNZ":
            #latest_version_of_schema = self.__latest_versions["CNZ"]
        #elif lixi_transaction_type == "DAS":
            #latest_version_of_schema = self.__latest_versions["DAS"]
        #elif lixi_transaction_type == "LMI":
            #latest_version_of_schema = self.__latest_versions["LMI"]
        #elif lixi_transaction_type == "SVC":
            #latest_version_of_schema = self.__latest_versions["SVC"]
        #elif lixi_transaction_type == "VAL":
            #latest_version_of_schema = self.__latest_versions["VAL"]
        #else:
            #raise LIXIInvalidSyntax(
                #"Schema read failed. Invalid transaction type provided."
            #)
        
        
        if lixi_transaction_type in self.__latest_versions:
            return self.__latest_versions[lixi_transaction_type]
        else:
            if self.schema_folder_path != None:
            
                version = self.__get_folder_latest_version__(lixi_transaction_type, latest_version_of_schema)
            
                if lixi_transaction_type not in self.__latest_versions:
                    self.__latest_versions[lixi_transaction_type] = version
            
                return version
            else:
                raise LIXIInvalidSyntax(
                    "Schema read failed. Schema folder path not set."
                )                
        #else:
            #return latest_version_of_schema

    def fetch_json_schema(
        self,
        lixi_transaction_type=None,
        lixi_version=None,
        custom_version=None,
        schema_string=None,
        schema_path=None,
        output_path=None,
        return_annotated = False,
        exploded = False        
    ):
        """Fetches a LIXI JSON schema dict object or converts from an LIXI XML schema.

        Args:
            lixi_transaction_type (:obj:`str`, optional): The transaction type of the LIXI schema to be fetched. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None.
            lixi_version (:obj:`str`, optional): The version of the LIXI schema to be fetched. Should be in the format of '2.6.24'. Defaults to None. Defaults to the latest version if only lixi_transaction_type is provided.
            custom_version (:obj:`str`, optional): The version of the LIXI custom schema to be fetched. Usually a complete file name of the custom schema. Defaults to None.
            schema_string (:obj:`str`, optional): LIXI schema provided as a string. Defaults to None.
            schema_path (:obj:`str`, optional): LIXI schema provided as a path. Defaults to None.
            output_path (:obj:`str`, optional): Path to write the fetched lixi schema to. Defaults to None.
            return_annotated (:obj:`boolean`, optional): Variable to indicated if the returned schema should be annotated or not.
            exploded (:obj:`boolean`, optional): Variable to indicate if the returned schema should be exploded or not.
            
            
        Result:
            A Python dict object of the schema or saved to the output folder specified. 
    
        Raises:
            LIXIInvalidSyntax: Incorrect schema string/file fetched.
            LIXIValidationError: Validation errors in the schema string/file fetched.
            LIXIResouceNotFoundError: File not found or can't be written.
        """
        # Make sure schema is annotated if exploded set to true
        if exploded==True:
            return_annotated = True
            
        # Determine which source is given
        output_schema = None

        if (
            lixi_transaction_type == None
            and custom_version == None
            and schema_string == None
            and schema_path == None
        ):
            raise LIXIInvalidSyntax(
                "Fetch schema failed. Fetch parameters not specified."
            )

        # Custom schema is done first to avoid fetching un custom schema based on transaction and version number
        if custom_version != None:
            output_xml_schema = self.__load_schema__(custom_version=custom_version)
            

        # Fetch based on transaction type and version
        elif lixi_version != None or lixi_transaction_type != None:
            if lixi_transaction_type == None:
                raise LIXIInvalidSyntax(
                    "Fetch schema failed. Transaction type parameter not specified."
                )

            if lixi_version == None:
                if self.schema_folder_path == None:
                    raise LIXIResouceNotFoundError(
                        "Please specify a Transaction LIXI Version or set the schema folder with lixi.set_schema_folder() to fetch the latest in the schema folder.\n"
                        + str(error_message)
                    )
                else:
                    lixi_version = self.get_schema_latest_version(lixi_transaction_type)

            output_xml_schema = self.__load_schema__(
                lixi_transaction_type=lixi_transaction_type, lixi_version=lixi_version, return_annotated=return_annotated
            )
            

        # Fetch based on string or path
        elif schema_string != None or schema_path != None:

            if schema_path != None:
                output_xml_schema = self.__load_schema__(schema_path=schema_path, return_annotated=return_annotated)
            else:
                output_xml_schema = self.__load_schema__(schema_string=schema_string, return_annotated=return_annotated)

            
            
        # Explode the schema if flag set to True
        if exploded==True:
            output_xml_schema = _explode_schema.main(output_xml_schema)
            
        # Convert into JSON
        output_schema = _jsonschema_functions.convert_to_json_schema(
            output_xml_schema
        )        
            
        # Output the schema to a file or return object
        if output_path == None:
            return _json.dumps(
                        output_schema, sort_keys=True, indent=4, ensure_ascii=False
                    )
        else:
            self.__write__(
                "fetched_json_schema_output.json",
                output_path,
                str(
                    _json.dumps(
                        output_schema, sort_keys=True, indent=4, ensure_ascii=False
                    )
                ).strip(),
                "Fetch schema failed. Can not store the schema at the specified folder.",
            )

            return output_schema

    def fetch_xml_schema(
        self,
        lixi_transaction_type=None,
        lixi_version=None,
        custom_version=None,
        schema_string=None,
        schema_path=None,
        output_path=None,
        return_annotated = False,
        exploded = False
    ):
        """Fetches a LIXI XML schema Etree object or converts from an LIXI JSON schema.

        Args:
            lixi_transaction_type (:obj:`str`, optional): The transaction type of the LIXI schema to be fetched. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None.
            lixi_version (:obj:`str`, optional): The version of the LIXI schema to be fetched. Should be in the format of '2.6.24'. Defaults to None. Defaults to the latest version if only lixi_transaction_type is provided.
            custom_version (:obj:`str`, optional): The version of the LIXI custom schema to be fetched. Usually a complete file name of the custom schema. Defaults to None.
            schema_string (:obj:`str`, optional): LIXI schema provided as a string. Defaults to None.
            schema_path (:obj:`str`, optional): LIXI schema provided as a path. Defaults to None.
            output_path (:obj:`str`, optional): Path to write the fetched lixi schema to. Defaults to None.
            return_annotated (:obj:`boolean`, optional): Variable to indicated if the returned schema should be annotated or not.
            exploded (:obj:`boolean`, optional): Variable to indicate if the returned schema should be exploded or not.
            
        Result:
            A Python Etree object of the schema or saved to the output folder specified. 
    
        Raises:
            LIXIInvalidSyntax: Incorrect schema string/file fetched.
            LIXIValidationError: Validation errors in the schema string/file fetched.
            LIXIResouceNotFoundError: File not found or can't be written.
        """
        
        # Make sure schema is annotated if exploded set to true
        if exploded==True:
            return_annotated = True
            
        # Determine which source is given
        output_schema = None

        if (
            lixi_transaction_type == None
            and custom_version == None            
            and schema_string == None
            and schema_path == None
        ):
            raise LIXIInvalidSyntax(
                "Fetch schema failed. Fetch parameters not specified."
            )

        # Custom schema is done first to avoid fetching un custom schema based on transaction and version number
        if custom_version != None:
            output_schema = self.__load_schema__(custom_version=custom_version)

        # Fetch based on transaction type and version
        elif lixi_version != None or lixi_transaction_type != None:
            if lixi_transaction_type == None:
                raise LIXIInvalidSyntax(
                    "Fetch schema failed. Transaction type parameter not specified."
                )

            if lixi_version == None:
                if self.schema_folder_path == None:
                    raise LIXIResouceNotFoundError(
                        "Please specify a Transaction LIXI Version or set the schema folder with lixi.set_schema_folder() to fetch the latest in the schema folder.\n"
                        + str(error_message)
                    )
                else:                
                    lixi_version = self.get_schema_latest_version(lixi_transaction_type)

            output_schema = self.__load_schema__(
                lixi_transaction_type=lixi_transaction_type, lixi_version=lixi_version, return_annotated=return_annotated
            )

        # Fetch based on string or path
        elif schema_string != None or schema_path != None:

            if schema_path != None:
                output_schema = self.__load_schema__(schema_path=schema_path, return_annotated=return_annotated)
            else:
                output_schema = self.__load_schema__(schema_string=schema_string, return_annotated=return_annotated)
        
        # Explode the schema if flag set to True
        if exploded==True:
            output_schema = _explode_schema.main(output_schema)
                
        # Output the schema to a file or return object
        if output_path == None:
            return _etree.tostring(output_schema, pretty_print=True).decode("utf-8")
        else:
            self.__write__(
                "fetched_xml_schema_output.xsd",
                output_path,
                str(
                    _etree.tostring(output_schema, pretty_print=True).decode("utf-8")
                ).strip(),
                "Fetch schema failed. Can not store the schema at the specified folder.",
            )

            return output_schema

    def get_schema_paths(
        self,
        lixi_transaction_type=None,
        lixi_version=None,
        file_type="xml",
        custom_version=None,
        schema_string=None,
        schema_path=None,
        exploded = False
    ):
        """Fetches all of the elements paths of the LIXI schema provided as a Python List of paths. 

        Args:
            lixi_transaction_type (:obj:`str`, optional): The transaction type of the LIXI schema to be fetched. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None.
            lixi_version (:obj:`str`, optional): The version of the LIXI schema to be fetched. Should be in the format of '2.6.24'. Defaults to None. Defaults to the latest version if only lixi_transaction_type is provided.
            custom_version (:obj:`str`, optional): The version of the LIXI custom schema to be fetched. Usually a complete file name of the custom schema. Defaults to None.
            schema_string (:obj:`str`, optional): LIXI schema provided as a string. Defaults to None.
            schema_path (:obj:`str`, optional): LIXI schema provided as a path. Defaults to None.
            exploded (:obj:`bool`, optional): Variable to indicate if the returned schema should be exploded or not. 
    
        Returns:
            A python list of all paths in the LIXI schema.
    
        Raises:
            LIXIInvalidSyntax: Validation errors for the lixi version or transaction type.
            LIXIResouceNotFoundError: If path provided does not exist.
            LIXIValidationError: If path provided does not exist.
        """

        # Get the latest version of the transactiontion schema if required.
        if lixi_transaction_type != None and lixi_version == None:
            
            if self.schema_folder_path == None:
                raise LIXIResouceNotFoundError(
                    "Please specify a Transaction LIXI Version or set the schema folder with lixi.set_schema_folder() to fetch the latest schema in the schema folder, to obtain schema paths.\n"
                    + str(error_message)
                )
            else:            
            
                lixi_version = str(
                    self.get_schema_latest_version(lixi_transaction_type)
                ).replace("_", ".")

        # Load the required version of the transaction schema.
        schema = self.__load_schema__(
            lixi_transaction_type,
            lixi_version,
            file_type,
            custom_version,
            schema_string,
            schema_path,
            True,
            True
        )

        # Generate schema paths.
        if exploded:
            return _path_functions.get_exploded_paths_for_schema_elements(schema)
        else:
            return _path_functions.get_paths_for_schema_elements(schema)

    def generate_custom_schema(
        self,
        instructions=None,
        instructions_path=None,
        elements_array=None,
        lixi_transaction_type=None,
        lixi_version=None,
        schema_string=None,
        schema_path=None,
        output_name=None,
        output_folder=None,
        output_type="xml",
        output_instructions=False,
    ):
        """Generates a custom LIXI schema based on the given customization instructions source. This produced schema is based on the project (https://standards.lixi.org.au/lixi2/CustomisationByRestriction) that derives a subschema that only uses sub set of elements available in the full LIXI schema. This produced schema is a valid LIXI schema.

        Args:
            instructions (:obj:`str`, optional): Customization instructions provided as a string. Defaults to None.
            instructions_file (:obj:`str`, optional): Customization instructions provided as a path. Defaults to None.
            elements_array (:obj:`str`, optional): Customization paths to keep in an array. Defaults to None.
            lixi_transaction_type (:obj:`str`, optional): The transaction type of the LIXI schema to be fetched. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None.
            lixi_version (:obj:`str`, optional): The version of the LIXI schema to be fetched. Should be in the format of '2.6.24'. Defaults to None. Defaults to the latest version if only lixi_transaction_type is provided.
            schema_string (:obj:`str`, optional): LIXI schema provided as a string. Defaults to None.
            schema_path (:obj:`str`, optional): LIXI schema provided as a path. Defaults to None.
            output_name (:obj:`str`, optional): String to set as a name of the custom schema. Defaults to None.
            output_folder (:obj:`str`, optional): Path to write the produced lixi schema to. Defaults to None.
            output_type (:obj:`str`, optional): The output type of the custom LIXI schema produced. Currently only accepts 'xml' and 'json'.
            output_instructions (:obj:`boolean`, optional): A flag to indicate whether a customization instructions need to be saved to a path. Must have the output folder path set. Defaults to False. 
        
        Returns:
            A customised LIXI schema as a string which can also be to the output folder specified.
        
        Raises:
            LIXIInvalidSyntax: Validation errors for the lixi version or transaction type.
            LIXIResouceNotFoundError: If path provided does not exist.
            LIXIValidationError: validation error in the customization instructions.
        """
        # Read the instructions file
        if instructions != None or instructions_path != None:

            if instructions_path != None:
                if _os.path.exists(instructions_path) == True:
                    instructions = _etree.parse(instructions_path)
                    instructions = instructions.getroot()
                    instructions = _etree.tostring(instructions).decode("utf-8")
                else:
                    raise LIXIResouceNotFoundError(
                        "Customised schema generation failed. Instructions file not found at the specified path."
                    )

            if (
                lixi_transaction_type == None
                and schema_string == None
                and schema_path == None
            ):
                transactiontype = re.search(
                    r'LIXITransactionType="([A-Z]*)"', instructions
                )
                try:
                    lixi_transaction_type = transactiontype.group(1)
                except:
                    raise LIXIInvalidSyntax(
                        "Customised schema generation failed. Instructions file does not have the transaction type specified."
                    )                    
                
            if lixi_transaction_type != None and lixi_version==None:
                if self.schema_folder_path == None:
                    raise LIXIResouceNotFoundError(
                        "Please specify a Transaction LIXI Version or set the schema folder with lixi.set_schema_folder() to fetch the latest schema in the schema folder, to obtain a customised schema.\n"
                        + str(error_message)
                    )
                else:                
                    lixi_version = str(
                        self.get_schema_latest_version(lixi_transaction_type)
                    ).replace("_", ".")

            # Has to be XML and has to be transaction hence custom is None
            schema = self.__load_schema__(
                lixi_transaction_type,
                lixi_version,
                "xml",
                None,
                schema_string,
                schema_path,
                True,
                True
            )

        # Read the csv text/file. Assupmtion for csv text is 'Package,Package.Content,..........'
        elif elements_array != None:
            
            if lixi_transaction_type != None and lixi_version==None:
                if self.schema_folder_path == None:
                    raise LIXIResouceNotFoundError(
                        "Please specify a Transaction LIXI Version or set the schema folder with lixi.set_schema_folder() to fetch the latest schema in the schema folder, to obtain a customised schema.\n"
                        + str(error_message)
                    )
                else:                  
                
                    lixi_version = str(
                        self.get_schema_latest_version(lixi_transaction_type)
                    ).replace("_", ".")            
            
            schema = self.__load_schema__(
                lixi_transaction_type,
                lixi_version,
                "xml",
                None,
                schema_string,
                schema_path,
                True,
                True
            )  # Has to be XML and has to be transaction hence custom is None
            
            if len(elements_array)>0:
                include_elements_list = elements_array
                include_elements_list = [elem.strip() for elem in elements_array]
            else:
                raise LIXIResouceNotFoundError(
                    "Array provided for customization is not populated. Please iclude paths in the array for schema customisation.\n"
                    + str(error_message)
                )                

            #if csv_path != None:
                #if _os.path.exists(csv_path) == True:
                    #f = _io.open(csv_path, mode="r", encoding="utf-8")
                    #csv_text = f.read()
                    #f.close()
                #else:
                    #raise LIXIResouceNotFoundError(
                        #"Customised schema generation failed. CSV file not found at the specified path."
                    #)

            ## Since CSV is a elements-to-include list, derive a instructions file based on that.
            #if len(csv_text.split("\n")) > 0:
                #temp_list = csv_text.split("\n")            
            #elif len(csv_text.split(",")) > 0:
                #temp_list = csv_text.replace("\n", "").split(",")
            
            #include_elements_list = [elem.strip() for elem in temp_list]

            message_paths = _path_functions.get_paths_for_elements(
                paths_list=include_elements_list, schema=schema
            )
            schema_paths = _path_functions.get_paths_for_schema_elements(schema)

            instructions = _path_functions.get_blacklist_paths_for_customization(
                self.current_transaction_type, message_paths, schema_paths
            )

        
        # Determine if we need the CustomisationJSON element.
        # It's needed for JSON schema output, but not for XML schema output.
        include_json = output_type == "json"

        # Generate the customised schema as string
        (
            customised_schema,
            output_name,
        ) = _customise_schema.execute_customisation_instructions(
            schema, instructions, output_name, include_customisation_json=include_json
        )
        
        # Generate customization instructions if flag set to true
        if output_instructions==True:
            if output_folder!=None:
                cust_output_path = _os.path.join(output_folder, output_name + ".xml")
            else:
                # When output_folder is not specified, we should not assume current working directory.
                # The user might just want the string returned.
                # To avoid writing files unexpectedly, we can raise an error or just skip writing.
                # For now, let's skip writing if output_folder is None.
                pass
                
            self.__write__(
                "customised_instructions_output.xsd",
                cust_output_path,
                instructions,
                "Customised schema instructions generation failed. Can not store the produced custom schema at the specified folder.",
            )                
        
        # Setting ouput type
        if output_type == "xml":
            output = customised_schema  ##xsd
        elif output_type == "json":
            output = _jsonschema_functions.convert_to_json_schema(
                _etree.fromstring(customised_schema)
            )
            output = _json.dumps(output, sort_keys=True, indent=4, ensure_ascii=False)

        # Output as required.
        if output_folder == None:
            return output
        else:
            if output_type == "xml":
                output_path = _os.path.join(output_folder, output_name + ".xsd")
            elif output_type == "json":
                output_path = _os.path.join(output_folder, output_name + ".json")

            self.__write__(
                "custom_schema_output.xsd",
                output_path,
                output,
                "Customised schema generation failed. Can not store the produced custom schema at the specified folder.",
            )

            return output

    def generate_schema_documentation(
        self,
        schema_string, 
        output_folder = None, 
        output_name = None,
        glossary_string = None
    ):
        """Generates a custom LIXI schema based on the given customization instructions source. This produced schema is based on the project (https://standards.lixi.org.au/lixi2/CustomisationByRestriction) that derives a subschema that only uses sub set of elements available in the full LIXI schema. This produced schema is a valid LIXI schema.

        Args:
            schema_string (:obj:`str`, required): LIXI schema provided as a string.
            output_folder (:obj:`str`, required): Path to write the produced lixi schema documentation to. If output_path is specified as None, a zip bytes file is returned.
            output_name (:obj:`str`, optional): Name of the sub-folder that contains the documentation. Defaults to None.      
            glossary_string (:obj:`str`, optional): LIXI Glossary update provided as a string. Defaults to None.
        
        Returns:
            Outputs the custom documentation for a schema to the specified folder. 
        
        Raises:
            LIXIInvalidSyntax: If the provided schema_string is not a string or is unable to be read.
            LIXIResouceNotFoundError: If path provided does not exist.
        """
        
        json_schema_string = None 
        
        if output_folder!=None:
            if _os.path.exists(output_folder) == False:
                raise LIXIResouceNotFoundError(
                    "Schema documentation creation failed. Schema folder path specified does not exist."
                )            
        
        if type(schema_string) != str:
            raise LIXIInvalidSyntax(
                "Schema documentation creation failed. Schema provided is not in a string format or can not be read."
            )            
        
        if '"$schema"' in schema_string:  ## is json
            json_schema_string = schema_string
            xml_schema_string = self.fetch_xml_schema(schema_string = schema_string, return_annotated=True)
            
            if output_name == None:
                sv = _json.loads(json_schema_string)
                
                if '@LIXICustomVersion' in sv['properties']['Package']['properties']['SchemaVersion']['properties']:
                    if 'enum' in sv['properties']['Package']['properties']['SchemaVersion']['properties']['@LIXICustomVersion']:
                        custom = sv['properties']['Package']['properties']['SchemaVersion']['properties']['@LIXICustomVersion']['enum'][0]
                        output_name = custom
                    else:
                        version = sv['properties']['Package']['properties']['SchemaVersion']['properties']['@LIXIVersion']['enum'][0]
                        trxn = sv['properties']['Package']['properties']['SchemaVersion']['properties']['@LIXITransactionType']['enum'][0]
                        output_name = str('LIXI-'+trxn+'-'+version.replace('.','_')+'-Annotated')                        
                else:
                    version = sv['properties']['Package']['properties']['SchemaVersion']['properties']['@LIXIVersion']['enum'][0]
                    trxn = sv['properties']['Package']['properties']['SchemaVersion']['properties']['@LIXITransactionType']['enum'][0]
                    output_name = str('LIXI-'+trxn+'-'+version.replace('.','_')+'-Annotated')
            
        elif "<xs:schema" in schema_string:
            xml_schema_string = schema_string
            
            if output_name == None:
                etr = _etree.fromstring(xml_schema_string)
                sv = etr.xpath('./xs:element/xs:complexType/xs:sequence/xs:element[@name="SchemaVersion"]', namespaces=ns)[0]        
                
                if len(sv.xpath('./xs:complexType/xs:attribute[@name="LIXICustomVersion"]', namespaces=ns))>0:
                    custom = sv.xpath('./xs:complexType/xs:attribute[@name="LIXICustomVersion"]', namespaces=ns)[0].attrib["fixed"]
                    output_name = custom
                else:
                    version = sv.xpath('./xs:complexType/xs:attribute[@name="LIXIVersion"]', namespaces=ns)[0].attrib["fixed"]
                    trxn = sv.xpath('./xs:complexType/xs:attribute[@name="LIXITransactionType"]', namespaces=ns)[0].attrib["fixed"]
                    output_name = str('LIXI-'+trxn+'-'+version.replace('.','_')+'-Annotated')
            
        else:
            raise LIXIInvalidSyntax(
                "Schema documentation creation failed. Schema provided can not be read."
            )
        
        if 'annotation="Full"' not in xml_schema_string: 
            raise LIXIInvalidSyntax(
                "Schema documentation creation failed. Schema provided is not fully annotated, as required."
            )        
        
        schema_doc_zip = None
        try:
            schema_doc_zip =  _schema_documentation.create_schema_documentation(xml_schema_string, json_schema_string, output_folder, output_name, glossary_string)
        except Exception as e:    
            raise LIXIInvalidSyntax(
                "Schema documentation creation failed.\n\nError Message:\n    "
                + str(e)
            )
        
        return schema_doc_zip 

#################
## ERROR CLASSES
#################
class LIXIResouceNotFoundError(Exception):
    def __init__(self, message):
        super().__init__("" + message)


class LIXIInvalidSyntax(Exception):
    def __init__(self, message):
        super().__init__("" + message)


class LIXIValidationError(Exception):
    def __init__(self, message, message_instance=None, error_log=None):

        super().__init__(message)
        self.message_instance = message_instance
        self.error_log = error_log



import os, glob, re, isodate, datetime, io, json
from decimal import *
from lxml import etree as elemTree
from xmljson import BadgerFish
from collections import OrderedDict
from operator import attrgetter

if __name__ == "_LIXI":
    import _explode_schema
else:
    from lixi import (
        _explode_schema
    )

ns = {
    "xs": "http://www.w3.org/2001/XMLSchema",
    "lx": "lixi.org.au/schema/appinfo_elements",
    "li": "lixi.org.au/schema/appinfo_instructions",
}

def build_types(annotated_schema_obj):
    simpletype_mappings = {}
    type_mappings = {
        "anyType":  "string",
        "anySimpleType":  "string",
        "string":  "string",
        "normalizedString":  "string",
        "token":  "string",
        "language":  "string",
        "Name":  "string",
        "NCName":  "string",
        "ID":  "id",
        "IDREF":  "idref",
        "IDREFS":  "idref",
        "ENTITY":  "string",
        "ENTITIES":  "string",
        "NMTOKEN":  "string",
        "NMTOKENS": "string",
        "boolean": "boolean",
        "base64Binary": "base64",
        "hexBinary": "string",
        "float": "number",
        "decimal": "number",
        "integer": "integer",
        "nonPositiveInteger": "integer",
        "negativeInteger": "integer",
        "long": "integer",
        "int": "integer",
        "short": "integer",
        "byte": "integer",
        "nonNegativeInteger": "integer",
        "unsignedLong": "integer",
        "unsignedInt": "integer",
        "unsignedShort": "integer",
        "unsignedByte": "integer",
        "positiveInteger": "integer",
        "double": "number",
        "anyURI": "string",
        "QName": "string",
        "NOTATION": "string",
        "duration": "string",
        "dateTime": "datetime",
        "date": "date",
        "time": "string",
        "gYearMonth": "string",
        "gYear": "year",
        "gMonthDay": "string",
        "gDay": "string",
        "gMonth": "string"
    }
    
    exploded_annotated_schema_obj = _explode_schema.main(annotated_schema_obj)
    
    elementsandatt_mappings = {}

    for element in exploded_annotated_schema_obj.iter():
        if "www.w3.org" in element.tag:
            if "simpleType" in element.tag:
                element_name = element.attrib.get("name")
                element_type = element.find("xs:restriction", ns).attrib["base"]
                if "xs:" in element_type:
                    element_type = element_type[3:]

                simpletype_mappings[element_name] = element_type

            elif "attribute" in element.tag:
                element_path = element.find('./xs:annotation/xs:appinfo/lx:path', ns).text
                element_type = element.attrib["type"]

                elementsandatt_mappings[element_path] = element_type    

            elif "element" in element.tag:
                element_path = element.find('./xs:annotation/xs:appinfo/lx:path', ns).text
                
                if 'maxOccurs' in element.attrib:
                    if element.attrib["maxOccurs"] == '1':
                        element_max = 'obj'
                    else:
                            element_max = 'arr'
                else:
                    element_max = 'obj'  

                elementsandatt_mappings[element_path] = element_max

    for k,v in elementsandatt_mappings.items():
        if v in simpletype_mappings:
            elementsandatt_mappings[k] = type_mappings[simpletype_mappings[v]]

    return elementsandatt_mappings        

def to_json(xml_etree, xsd_annotated_schema):
    """ Converts a Python XML etree to JSON etree and arranges the result per schema.
    
    Args:
        xml_etree (:obj:`lxml.etree`, required): XML Message to convert to JSON message.
        xsd_annotated_schema (:obj:`lxml.etree`, required): A LIXI XML Schema.
    
    Returns:
        A python JSON dict object.
    
    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    json_object = {}

    def json_change_types(json_obj, parent_path, mappings):
        
        for key ,value in json_obj.items():

            this_elem_path = None
            if parent_path=='':
                this_elem_path = key
            else:
                this_elem_path = parent_path +'.'+ key
        
            if isinstance(value, list):
                for array_item in value:
                    json_change_types(array_item, this_elem_path, mappings)
                    
            elif isinstance(value, dict):
                if this_elem_path in mappings:
                    if mappings[this_elem_path]=='arr':
                        json_obj[key] = [value]
                        
                        for array_item2 in json_obj[key]:
                            json_change_types(array_item2, this_elem_path, mappings) 
                    else:
                        json_change_types(value, this_elem_path, mappings)
                else:
                    json_change_types(value, this_elem_path, mappings)
            else:
                this_elem_path = this_elem_path.replace('@','')
                if this_elem_path in mappings:
                    lookup = mappings[this_elem_path]
    
                    if "number" == lookup or "decimal" == lookup or "float" == lookup:
                        json_obj[key] = float(value)
                    elif "integer" == lookup:
                        json_obj[key] = int(value)

        return json_obj

    bf = BadgerFish(xml_fromstring=False)
    json_object = bf.data(xml_etree)

    mappings = build_types(xsd_annotated_schema)

    json_object = json_change_types(json_object, '', mappings)
      
    return json_object


def to_json_string(xml_str, xsd_annotated_schema_str):
    """ Converts a Python XML string to JSON string and arranges the result per schema.
    
    Args:
        xml_str (:obj:`str`, required): XML Message string to convert to JSON message.
        xsd_annotated_schema_str (:obj:`str`, required): A LIXI XML Schema string.
    
    Returns:
        A python JSON string message object.
    
    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    xml_etree = elemTree.fromstring(xml_str)
    xsd_annotated_schema = elemTree.fromstring(xsd_annotated_schema_str)

    converted_json = to_json(xml_etree, xsd_annotated_schema)
    return json.dumps(converted_json, sort_keys=True, indent=4, ensure_ascii=False)


def to_xml(json_dict, xsd_annotated_schema):
    """ Converts a Python JSON dict to XML etree and arranges the result per schema.
    
    Args:
        json_dict (:obj:`dict`, required): JSON Message string to convert to XML message.
        xsd_annotated_schema (:obj:`lxml.etree`, required): A LIXI XML Schema string.
    
    Returns:
        A python XML Etree object.
    
    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    def xml_get_sorted_list(schema_root):
        sorting_elem_list = {}
        sorting_list = {}
        sorting_complex_list = {}

        for elem in schema_root.iter(tag=("{*}element", "{*}complexType")):

            if "element" in elem.tag and elem.attrib.get("name") is not None:
                if elem.attrib.get("type") is None:
                    key_name = (
                        str(
                            elem.xpath(
                                "./xs:annotation/xs:appinfo/lx:path/text()",
                                namespaces=ns,
                            )
                        )
                        .strip("[]'")
                        .replace(".", "/")
                    )

                    if (
                        elem.find(
                            "./xs:complexType/xs:sequence/xs:element", namespaces=ns
                        )
                        is not None
                    ):
                        sorting_elem_list[key_name] = []
                        for child_xsd_element in elem.xpath(
                            "./xs:complexType/xs:sequence/xs:element", namespaces=ns
                        ):
                            sorting_elem_list[key_name].append(
                                child_xsd_element.attrib.get("name")
                            )

                    elif (
                        elem.find(
                            "./xs:complexType/xs:choice/xs:element", namespaces=ns
                        )
                        is not None
                    ):
                        sorting_elem_list[key_name] = []
                        for child_xsd_element in elem.findall(
                            "./xs:complexType/xs:choice/xs:element", namespaces=ns
                        ):
                            sorting_elem_list[key_name].append(
                                child_xsd_element.attrib.get("name")
                            )

                elif elem.attrib.get("type") is not None:
                    key_name = (
                        str(
                            elem.xpath(
                                "./xs:annotation/xs:appinfo/lx:path/text()",
                                namespaces=ns,
                            )
                        )
                        .strip("[]'")
                        .replace(".", "/")
                    )
                    sorting_elem_list[key_name + "/" + elem.attrib.get("type")] = []

            elif "complexType" in elem.tag:
                if elem.attrib.get("name") is not None:
                    key_name = (
                        str(
                            elem.xpath(
                                "./xs:annotation/xs:appinfo/lx:path/text()",
                                namespaces=ns,
                            )
                        )
                        .strip("[]'")
                        .replace(".", "/")
                    )

                    if elem.find("./xs:sequence/xs:element", namespaces=ns) is not None:
                        sorting_complex_list[key_name] = []
                        for child_xsd_element in elem.xpath(
                            "./xs:sequence/xs:element", namespaces=ns
                        ):
                            sorting_complex_list[key_name].append(
                                child_xsd_element.attrib.get("name")
                            )

        for k, v in sorting_complex_list.items():
            for key, value in sorting_elem_list.items():
                if k in key:
                    sorting_list["/" + key.replace("/" + k, "")] = v

        for k, v in sorting_elem_list.items():
            if len(v) > 0:
                sorting_list["/" + k] = v

        return sorting_list

    def xml_sort_message(unsorted_xml, sorting_list):

        tree = elemTree.ElementTree(unsorted_xml)

        for e in unsorted_xml.xpath("//*"):

            keys = e.attrib
            sorted_attrib = OrderedDict(sorted(keys.items()))
            e.attrib.clear()
            e.attrib.update(sorted_attrib)

            elem_path_key = tree.getpath(e).strip("[0123456789]")

            if elem_path_key in sorting_list:
                e[:] = sorted(
                    e, key=lambda elem: get_elem_key(elem, elem_path_key, sorting_list)
                )

        return unsorted_xml

    def get_elem_key(elem, elem_path_key, sorting_list):

        try:
            if sorting_list[elem_path_key].index(elem.tag) is not None:
                return sorting_list[elem_path_key].index(elem.tag)
            else:
                return 0
        except:
            return 0

    bf = BadgerFish()
    unsorted_xml = bf.etree(json_dict)[0]

    sorting_list = xml_get_sorted_list(xsd_annotated_schema)
    converted_sorted_xml = xml_sort_message(unsorted_xml, sorting_list)

    return converted_sorted_xml


def to_xml_string(json_str, xsd_annotated_schema_str):
    """ Converts a Python JSON string to XML string and arranges the result per schema.
    
    Args:
        json_str (:obj:`str`, required): JSON Message string to convert to XML message.
        xsd_annotated_schema_str (:obj:`str`, required): A LIXI XML Schema string.
    
    Returns:
        A python XML string message object.
    
    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    json_dict = json.loads(json_str)
    xsd_annotated_schema = elemTree.fromstring(xsd_annotated_schema_str)

    converted_xml = to_xml(json_dict, xsd_annotated_schema)
    return elemTree.tostring(converted_xml, pretty_print=True).decode("utf-8")



"""Setup script for lixi2 package project"""

import os, os.path
from setuptools import setup, find_packages

# The directory containing this file
HERE = os.path.abspath(os.path.dirname(__file__))

# The text of the README file
with open(os.path.join(HERE, "README.md")) as fid:
    README = fid.read()


# This call to setup() does all the work
setup(
    name="lixi",
    version= str(os.environ.get('test_version')),
    description="lixi is a python package to manipulate a valid LIXI 2 message and schema.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://standards.lixi.org.au/lixi-tech/lixi-pypi",
    author="Ammar Khan",
    author_email="lixilab@lixi.org.au",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(exclude=['build','dist','lixi.egg-info','lixi_ammar.egg-info','tests','.git', 'docs']),
    include_package_data=True,
    install_requires=['lxml>=4.4.1','isodate>=0.6.0','xmljson>=0.2.0','jsonschema>=3.0.2','importlib_resources>=6.4.0'] 
)
